package fat-filesystem

  1. Overview
  2. Docs

Pure OCaml implementation of the FAT filesystem

This library has two purposes:

  1. to allow the easy preparation of bootable disk images containing mirage kernels
  2. to provide a simple filesystem layer for mirage applications